"""Functions for preprocessing ULFs."""

from functools import reduce

from transduction import tt
from memoization import cached

from ulflib.util import atom, append, cons, flatten_singletons, listp, apply_all, new_var, variables_in
from ulflib.ulflib import (direct_sent_mod_stop_p, arg_sent_mod_stop_p, sent_mod_stop_p, sent_mod_p, ps_p, make_all_explicit, lower_all,
                            split_by_suffix, apply_substitution_macros,
                            lex_adv_e_p, adv_e_p, lex_adv_f_p, adv_f_p, lex_number_p, pp_p, lex_pronoun_p, lex_det_p, adj_p, lex_adjective_p,
                            lex_mod_n_p, mod_a_p)
from ulflib import scoping
from ulflib.tense_tree.temporal_lex import *

# ``````````````````````````````````````
# TT Temporal Adverb LF Patterns
# ``````````````````````````````````````


@cached
def number_p(x):
  return lex_number_p(x) or x in NUMBERS

@cached
def to_number(x):
  return int(x) if lex_number_p(x) else NUMBERS[x]

@cached
def adv_e_pp(x):
  return adv_e_p(x) and listp(x) and len(x) == 2 and pp_p(x[1])

@cached
def adv_e_adj(x):
  return adv_e_p(x) and listp(x) and len(x) == 2 and adj_p(x[1])

@cached
def adv_f_pp(x):
  return adv_f_p(x) and listp(x) and len(x) == 2 and pp_p(x[1])

@cached
def adv_f_adj(x):
  return adv_f_p(x) and listp(x) and len(x) == 2 and adj_p(x[1])


@cached
def ever_adv(x, oldvars):
  if x == 'ever.adv-e':
    return ['adv-e', 'episode.n'], []
  else:
    return None, []


@cached
def indexical_adv(x, oldvars):
  """Yesterday, now, etc."""
  ret = None
  newvars = []
  if x == 'currently.adv-e' or x == ['adv-e', 'current.a']:
    ret = ['adv-e', ['during.p', '^now']]
  elif lex_adv_e_p(x):
    word, _ = split_by_suffix(x)
    if word in INDEXICAL_TIMES:
      ret = ['adv-e', ['during.p', '^'+word]]
  return ret, newvars


@cached
def temp_location_adv(x, oldvars):
  """On Friday, during last weekend, etc."""
  ret = None
  newvars = []
  if lex_adv_e_p(x):
    word, _ = split_by_suffix(x)
    if word in LOCATION_ADV_MAP:
      adj = LOCATION_ADV_MAP[word]
      var = new_var(oldvars, base='?t')
      ret = ['adv-e', ['during.p', ['the', var, [var, [adj, 'episode.n']]]]]
      newvars = [var]
  elif adv_e_adj(x):
    if lex_adjective_p(x[1]):
      lex_adj, _ = split_by_suffix(x[1])
      if lex_adj in LOCATION_ADJ_MAP:
        adj = LOCATION_ADJ_MAP[lex_adj]
        var = new_var(oldvars, base='?t')
        ret = ['adv-e', ['during.p', ['the', var, [var, [adj, 'episode.n']]]]]
        newvars = [var]
  elif adv_e_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in LOCATION_PRED_MAP:
      pred = LOCATION_PRED_MAP[lex_prep]
      # Indexical term, e.g. During Friday
      if atom(termp):
        lex_term, _ = split_by_suffix(termp)
        if termp in INDEXICAL_TIMES or (lex_pronoun_p(termp) and lex_term in INDEXICAL_TIMES):
          ret = ['adv-e', [pred, '^'+lex_term]]
      # Noun phrase, e.g. During {the} last weekend
      else:
        det, np = termp
        if det in QUAN_MAP:
          quan = QUAN_MAP[det]
          var = new_var(oldvars, base='?t')
          ret = ['adv-e', [pred, [quan, var, [var, np]]]]
          newvars = [var]
  return ret, newvars


@cached
def temp_range_adv(x, oldvars):
  """Recently, before, next, etc."""
  ret = None
  newvars = []
  if lex_adv_e_p(x):
    word, _ = split_by_suffix(x)
    if word in RANGE_ADV_MAP:
      adj = RANGE_ADV_MAP[word]
      var = new_var(oldvars, base='?t')
      ret = ['adv-e', ['during.p', ['some', var, [var, [adj, 'episode.n']]]]]
      newvars = [var]
  elif adv_e_adj(x):
    if lex_adjective_p(x[1]):
      lex_adj, _ = split_by_suffix(x[1])
      if lex_adj in RANGE_ADJ_MAP:
        adj = RANGE_ADJ_MAP[lex_adj]
        var = new_var(oldvars, base='?t')
        ret = ['adv-e', ['during.p', ['some', var, [var, [adj, 'episode.n']]]]]
        newvars = [var]
    elif listp(x[1]) and mod_a_p(x[1][0]) and x[1][1] == 'ago.a':
      adj = x[1]
      var = new_var(oldvars, base='?t')
      ret = ['adv-e', ['during.p', ['some', var, [var, [adj, 'episode.n']]]]]
      newvars = [var]
  return ret, newvars


@cached
def duration_adv(x, oldvars):
  """For two hours, for three days, in three days, etc."""
  ret = None
  newvars = []
  if adv_e_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in DURATION_PRED_MAP and listp(termp):
      pred = DURATION_PRED_MAP[lex_prep]
      det, np = termp
      if lex_det_p(det):
        lex_det, _ = split_by_suffix(det)
        if number_p(lex_det):
          ret = ['adv-e', [pred, ['k', [str(to_number(lex_det))+'.mod-n', np]]]]
  return ret, newvars


@cached
def count_adv(x, oldvars):
  """Twice, {for} three times, etc."""
  ret = None
  newvars = []
  if lex_adv_f_p(x):
    word, _ = split_by_suffix(x)
    if number_p(word):
      ret = ['adv-f', [str(to_number(word))+'.mod-n', ['plur', 'episode.n']]]
  elif adv_f_adj(x):
    if lex_adjective_p(x[1]):
      lex_adj, _ = split_by_suffix(x[1])
      if number_p(lex_adj):
        ret = ['adv-f', [str(to_number(lex_adj))+'.mod-n', ['plur', 'episode.n']]]
  elif adv_f_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in ADV_F_P and listp(termp):
      det, np = termp
      if lex_det_p(det):
        lex_det, _ = split_by_suffix(det)
        if number_p(lex_det):
          ret = ['adv-f', [str(to_number(lex_det))+'.mod-n', np]]
  return ret, newvars


@cached
def frequency_adv(x, oldvars):
  """Frequently, {for} many times, rarely, usually, etc."""
  ret = None
  newvars = []
  if lex_adv_f_p(x):
    word, _ = split_by_suffix(x)
    if word in FREQUENCY_ADV_MAP:
      adj = FREQUENCY_ADV_MAP[word]
      var = new_var(oldvars, base='?t')
      ret = ['adv-f', [['attr', adj], ['plur', 'episode.n']]]
  elif adv_f_adj(x):
    if lex_adjective_p(x[1]):
      lex_adj, _ = split_by_suffix(x[1])
      if lex_adj in FREQUENCY_ADJ_MAP:
        adj = FREQUENCY_ADJ_MAP[lex_adj]
        ret = ['adv-f', [['attr', adj], ['plur', 'episode.n']]]
  elif adv_f_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in ADV_F_P and listp(termp):
      det, np = termp
      if lex_det_p(det):
        lex_det, _ = split_by_suffix(det)
        if lex_det in FREQUENCY_DET_MAP:
          adj = FREQUENCY_DET_MAP[lex_det]
          ret = ['adv-f', [['attr', adj], ['plur', 'episode.n']]]
  return ret, newvars


@cached
def periodicity_adv(x, oldvars):
  """{For} every four hours, etc."""
  ret = None
  newvars = []
  if adv_f_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in ADV_F_P and listp(termp):
      det, np = termp
      if lex_det_p(det):
        lex_det, _ = split_by_suffix(det)
        if lex_det in ['every', 'each']:
          var = new_var(oldvars, base='?s')
          if listp(np) and len(np) == 2 and (lex_adjective_p(np[0]) or lex_mod_n_p(np[0])) and number_p(split_by_suffix(np[0])[0]):
            mod = str(to_number(split_by_suffix(np[0])[0])) + '.mod-n'
            ret = ['adv-f', [':l', var, [[var, [['attr', 'periodic.a'], ['plur', 'episode.n']]], 'and.cc',
                                         [['period-of.f', '?s'], '=', ['k', [mod, np[1]]]]]]]
            newvars = [var]
  return ret, newvars


@cached
def recurrence_adv(x, oldvars):
  """{For} every Saturday, always, etc."""
  ret = None
  newvars = []
  if lex_adv_f_p(x):
    word, _ = split_by_suffix(x)
    if word in RECURRENCE_ADV_MAP:
      np = RECURRENCE_ADV_MAP[word]
      var_s = new_var(oldvars, base='?s')
      var_t = new_var(oldvars+[var_s], base='?t')
      var_e = new_var(oldvars+[var_s, var_t], base='?e')
      ret = ['adv-f', [':l', var_s, ['all', var_t, [var_t, np],
                                      ['exists', var_e, [[var_e, 'member-of.p', var_s], 'and.cc',
                                                        [var_e, 'during.p', var_t]]]]]]
  elif adv_f_adj(x):
    if lex_adjective_p(x[1]):
      lex_adj, _ = split_by_suffix(x[1])
      if lex_adj in RECURRENCE_ADJ_MAP:
        np = RECURRENCE_ADJ_MAP[lex_adj]
        var_s = new_var(oldvars, base='?s')
        var_t = new_var(oldvars+[var_s], base='?t')
        var_e = new_var(oldvars+[var_s, var_t], base='?e')
        ret = ['adv-f', [':l', var_s, ['all', var_t, [var_t, np],
                                        ['exists', var_e, [[var_e, 'member-of.p', var_s], 'and.cc',
                                                          [var_e, 'during.p', var_t]]]]]]
  elif adv_f_pp(x):
    pp = x[1]
    lex_prep, _ = split_by_suffix(pp[0])
    termp = pp[1]
    if lex_prep in ADV_F_P and listp(termp):
      det, np = termp
      if lex_det_p(det):
        lex_det, _ = split_by_suffix(det)
        if lex_det in ['every', 'each']:
          var_s = new_var(oldvars, base='?s')
          var_t = new_var(oldvars+[var_s], base='?t')
          var_e = new_var(oldvars+[var_s, var_t], base='?e')
          ret = ['adv-f', [':l', var_s, ['all', var_t, [var_t, np],
                                         ['exists', var_e, [[var_e, 'member-of.p', var_s], 'and.cc',
                                                            [var_e, 'during.p', var_t]]]]]]
          newvars = [var_s, var_t, var_e]
  return ret, newvars


ADV_E_MAPS = [ever_adv, indexical_adv, temp_location_adv, temp_range_adv, duration_adv]
ADV_F_MAPS = [count_adv, frequency_adv, periodicity_adv, recurrence_adv]


TEMPORAL_ADV_PREDS = [

]

for pred in TEMPORAL_ADV_PREDS:
  tt.register_pred(pred, include_neg=True)



# ``````````````````````````````````````
# Main Preprocessing Functions
# ``````````````````````````````````````



def uninvert_aux(ulf):
  """Uninvert any auxiliary verb + NP construction.
  
  E.g., ``((pres verb.aux-s) |A| ...)`` -> ``(|A| ((pres verb.aux-s) ...))``
  """
  return tt.apply_rules([(
    ['!tensed-aux-p', '!term-p', '+expr'],
    ['2', ['1', '3']])],
  ulf)


def lift_sent_mods(ulf):
  """Lift any sentential modifiers to the appropriate non-floating level, preventing lifting through particular lift-stopping operators."""
  def add_mods(expr, mods):
    mods = handle_sentential_mods(mods)
    return reduce(lambda x,y: [y,x], mods[::-1], expr)
  
  def split_mods(form):
    nomod = [x for x in form if not sent_mod_p(x)]
    mods = [x for x in form if sent_mod_p(x)]
    return nomod, mods
  
  def handle_sentential_mods(mods):
    return [[m[0], lift_sent_mods(m[1])] if ps_p(m) else m for m in mods]
  
  def rec(form):
    # Base case
    if atom(form):
      return form, []
    # Simple recursive case (no lift-stopping and no sent mods)
    # Recurse and return the removed ULFs together and append modifier lists
    elif not sent_mod_stop_p(form[0]) and not any([sent_mod_p(x) for x in form]):
      recres = [rec(x) for x in form]
      newulf = [r[0] for r in recres]
      allmods = append([r[1] for r in recres])
      return newulf, allmods
    # Simple-ish recursive case (no lift-stopping but has modifiers)
    # Recurse and remove current modifiers
    elif not sent_mod_stop_p(form[0]):
      nomod, mods = split_mods(form)
      reculf, recmods = rec(nomod)
      return reculf, mods+recmods
    # Direct lift-stopping
    elif direct_sent_mod_stop_p(form[0]):
      nomod, mods = split_mods(form)
      recres = [rec(x) for x in nomod]
      newulf = [r[0] for r in recres]
      allmods = mods + append([r[1] for r in recres])
      if len(newulf) != 2:
        raise Exception('All direct not stopping operators must have a single argument: ' + str(newulf))
      return [newulf[0], add_mods(newulf[1], allmods)], []
    # Argument lift-stopping
    # NB: sentential mods not in the arguments are lifted
    else:
      nomod, mods = split_mods(form)
      recres = [rec(x) for x in nomod]
      modadded_args = [add_mods(argres[0], argres[1]) for argres in recres[1:]]
      return cons(form[0], modadded_args), mods
  
  # Main function body
  # Recurse and add in any stray modifiers
  newulf, mods = rec(ulf)
  return flatten_singletons(add_mods(newulf, mods))


def expand_temporal_mods(ulf):
  """Expand any temporal modifiers to the full logical form.
  
  E.g., ``twice.adv-f`` -> ``(adv-f (for.p (k (two.a (plur episode.n)))))``
  """
  def rec(x, vars):
    if adv_e_p(x):
      new, newvars = apply_all((x, vars), ADV_E_MAPS, return_size=2)
      newvars = [] if newvars is None else newvars
      return (x,vars) if new is None else (new,vars+newvars)
    elif adv_f_p(x):
      new, newvars = apply_all((x, vars), ADV_F_MAPS, return_size=2)
      newvars = [] if newvars is None else newvars
      return (x,vars) if new is None else (new,vars+newvars)
    elif atom(x):
      return (x,vars)
    else:
      ret = []
      for y in x:
        part, partvars = rec(y, vars)
        ret.append(part)
        vars += partvars
      return (ret,vars)
  ret, _ = rec(ulf, variables_in(ulf))
  return ret


def preprocess(ulf, is_scoped=False):
  """Preprocess a ulf, including scoping tense operators (if necessary) and lifting/expanding temporal modifiers."""
  ulf = make_all_explicit(lower_all(ulf))
  ulf = apply_substitution_macros(ulf)
  ulf = uninvert_aux(ulf)
  if not is_scoped:
    ulf = scoping.scope(ulf, types=['tense'])
  return expand_temporal_mods(lift_sent_mods(ulf))


import sys
sys.path.append("src/")

import pytest

from ulflib import scoping


class TestScoping:
  """Basic unit tests for scoping algorithm.
  
  Notes
  -----
  TODO: these unit tests need to be validated and expanded further.
  """
  def test_1(self):
    ulf = ['John.name', [['past', 'tell.v'], 'Mary.name', 'something.pro']]
    scoped = ['past', ['John.name', ['tell.v', 'Mary.name', 'something.pro']]]
    assert scoping.scope(ulf) == scoped

  def test_2(self):
    ulf = ['John.name', [['past', 'tell.v'], 'Mary.name', ['that', ['he.pro', [['pres', 'go.v'], ['to.p', ['the.d', 'store.n']]]]]]]
    scoped = ['past', ['John.name', ['tell.v', 'Mary.name', ['that', ['pres', ['the', '?x', ['?x', 'store.n'], ['he.pro', ['go.v', ['to.p', '?x']]]]]]]]]
    assert scoping.scope(ulf) == scoped

  def test_3(self):
    ulf = ['she.pro', [['past', 'perf'], [['pasv', 'give.v'], ['an.d', 'award.n']]]]
    scoped = [['past', 'perf'], ['a{n}', '?x', ['?x', 'award.n'], ['she.pro', [['pasv', 'give.v'], '?x']]]]
    assert scoping.scope(ulf) == scoped

  def test_4(self):
    ulf = ['Mary.name', [['pres', 'prog'], ['try.v', ['to', ['ignore.v', ['an.d', 'itch.n']]]]]]
    scoped = [['pres', 'prog'], ['a{n}', '?x', ['?x', 'itch.n'], ['Mary.name', ['try.v', ['to', ['ignore.v', '?x']]]]]]
    assert scoping.scope(ulf) == scoped

  def test_5(self):
    ulf = ['John.name', [['past', 'tell.v'], 'Mary.name',
                         ['that', ['he.pro', [['pres', 'go.v'], ['to.p', ['the.d',
                                                [':l', '?x', [['?x', 'store.n'], 'and.cc', ['?x', ['near.p', ['the.d', 'corner.n']]]]]]]]]]]]
    scoped = ['past', ['John.name', ['tell.v', 'Mary.name',
                                     ['that', ['pres', ['the', '?x1', [['?x1', 'store.n'], 'and.cc',
                                                          ['the', '?x12', ['?x12', 'corner.n'], ['?x1', ['near.p', '?x12']]]],
                                                            ['he.pro', ['go.v', ['to.p', '?x1']]]]]]]]]
    assert scoping.scope(ulf) == scoped
    
    
class TestOptionalKeywords():
  def test_1(self):
    ulf = ['John.name', [['past', 'tell.v'], 'Mary.name', ['that', ['he.pro', [['pres', 'go.v'], ['to.p', ['the.d', 'store.n']]]]]]]
    scoped = ['past', ['John.name', ['tell.v', 'Mary.name', ['that', ['pres', ['he.pro', ['go.v', ['to.p', ['the.d', 'store.n']]]]]]]]]
    assert scoping.scope(ulf, types=['tense']) == scoped

  def test_2(self):
    ulf = ['John.name', [['past', 'tell.v'], 'Mary.name', ['that', ['he.pro', [['pres', 'go.v'], ['to.p', ['the.d', 'store.n']]]]]]]
    scoped = ['the', '?x', ['?x', 'store.n'], ['John.name', [['past', 'tell.v'], 'Mary.name', ['that', ['he.pro', [['pres', 'go.v'], ['to.p', '?x']]]]]]]
    print(scoping.scope(ulf, types=['quan']))
    assert scoping.scope(ulf, types=['quan']) == scoped


"""Tense tree implementation"""

from ulflib.util import atom, listp
from ulflib.sexpr import parse_s_expr
from ulflib.ulflib import lex_tense_p, lex_aspect_p, tensed_sent_reifier_p, adv_e_p, adv_f_p, lex_rel_p, ps_p

from ulflib.tense_tree.constants import *
from ulflib.tense_tree.preprocess import preprocess
from ulflib.tense_tree.tense_node import TenseNode

def get_mood(ulf):
  """Determine the mood of the ULF."""
  if atom(ulf):
    return MOOD_DECL
  elif ulf[-1] in ['?', '.?']:
    return MOOD_QUES
  elif ulf[-1] == '!':
    return MOOD_IMPER
  else:
    return MOOD_DECL
  

class TenseTree:
  """A tense tree.
  
  Attributes
  ----------
  root : TenseNode
    The root node of this tense tree.
  now_idx : int
    A counter for the current 'Now' episode, to be updated
    with each logical form that gets deindexed.
  ep_idx : int
    A counter for each episode variable created.
  utt_idx : int
    A counter for each utterance variable created.
  disable_speech_acts : bool
    Whether to globally set speech_act = False when deindexing.
  """

  def __init__(self, disable_speech_acts=False):
    self.reset()
    self.disable_speech_acts=disable_speech_acts


  def reset(self):
    self.now_idx = 0
    self.ep_idx = 0
    self.utt_idx = 0
    self.root = TenseNode()


  def now(self):
    """Return the current 'Now' symbol."""
    return NOW + str(self.now_idx)


  def new_now(self):
    """Create a new 'Now' symbol."""
    self.now_idx += 1
    return self.now()


  def new_ep_var(self):
    """Create a new episode variable."""
    self.ep_idx += 1
    return EP_VAR + str(self.ep_idx)
  

  def new_utt_var(self):
    """Create a new utterance variable."""
    self.utt_idx += 1
    return UTT_VAR + str(self.utt_idx)
  

  def get_emb_episode(self, emb_node):
    """Get the episode corresponding to an embedding node.
    
    This is either the last episode of the node, or the current 'Now' if the node
    is the root of the tense tree.
    """
    if emb_node is self.root:
      return self.now()
    else:
      return emb_node.last_episode()
    

  def deindex(self, ulf, speech_act=True, is_scoped=False):
    """The top-level function to de-index a (possibly unscoped) logical form.

    Parameters
    ----------
    ulf : s-expr
      An "S-expression" (possibly nested lists of strings) representing an
      unscoped logical form (ULF) formula.
    speech_act : bool, default=True
      Whether to treat the input formula as a speech act, i.e., adding top-level
      speech act episodes to the resulting formula.
    is_scoped : bool, default=False
      Whether the input formula has already been scoped.
    
    Notes
    -----
    TODO: I'm not currently sure what should be done in the case of combined tense/aspect operators,
    such as (pres perf), or even whether this is currently supported by the scoping method used in
    preprocessing ULFs.
    """
    def deindex_recur(ulf, emb_node, focus):
      assert isinstance(emb_node, TenseNode) and isinstance(focus, TenseNode)
      
      if atom(ulf):
        return ulf
      op = ulf[0]

      # 1. Tense or aspect operator
      if lex_tense_p(op) or lex_aspect_p(op):
        e = self.new_ep_var()
        e_emb = self.get_emb_episode(emb_node)

        if op == PAST:
          restrictor = [e, focus.bef(), e_emb]
          new_focus = focus.move_left()
        elif op == PRES:
          restrictor = [e, AT_ABOUT, e_emb]
          new_focus = focus
        elif op == FUTR:
          restrictor = [e, AFTER, focus.last_episode()]
          new_focus = focus.move_right()
        elif op == PERF:
          restrictor = [e, IMPINGES_ON, focus.last_episode()]
          new_focus = focus.move_down()
        # other tense/aspect operators such as 'prog' or 'cf' are not yet supported
        else:
          return [op, deindex_recur(ulf[1], emb_node, focus)]
        
        if new_focus.last_episode():
          restrictor = [restrictor, AND, [new_focus.last_episode(), ORIENTS, e]]

        nucleus = [deindex_recur(ulf[1], emb_node, new_focus.add_episode(e)), CH, e]
        return [EXISTS, e, restrictor, nucleus]
      
      # 2. Embedding operator
      elif tensed_sent_reifier_p(op):
        return [op, deindex_recur(ulf[1], focus, focus.move_embedded_new())]

      # 3. Temporal or frequency adverb
      elif (adv_e_p(op) or adv_f_p(op)) and listp(op):
        e = focus.last_episode()
        adv_pred = deindex_recur(op[1], emb_node, focus)
        nucleus = deindex_recur(ulf[1], emb_node, focus)
        if adv_e_p(op):
          return [[e, adv_pred], AND, nucleus]
        else:
          return [[e, adv_pred], AND, [MULT, nucleus]]

      # 4. Relativizers (e.g., that.rel)
      elif lex_rel_p(op):
        # TODO
        pass

      # 5. Sentential prepositions (e.g., before.ps)
      elif ps_p(op) and listp(op):
        e = focus.last_episode()
        ps_res = deindex_recur(op[1], emb_node, focus)
        prep = op[0][:-1]
        if ps_res[0] == EXISTS:
          _, e2, restrictor, nucleus = ps_res
          restrictor_new = restrictor + [AND, [e, prep, e2]]
          return [EXISTS, e2, restrictor_new, nucleus]

      return [deindex_recur(ulf_part, emb_node, focus) for ulf_part in ulf]

    # ===== MAIN BODY OF FUNCTION =====
    
    if isinstance(ulf, str):
      ulf = parse_s_expr(ulf)

    mood = get_mood(ulf)
    ulf = preprocess(ulf, is_scoped=is_scoped)
    u = self.new_utt_var()
    now = self.new_now()
    restrictor = [u, SAME_TIME, now]
    focus = self.root

    # The utterance episode is assumed to be just after the preceeding episode in focus
    if focus.last_episode():
      restrictor = [restrictor, AND, [u, JUST_AFTER, focus.last_episode()]]

    embedded = deindex_recur(ulf, focus, focus.add_episode(u).move_embedded_cur())

    if self.disable_speech_acts or not speech_act:
      return embedded

    # Form speech act logical form depending on mood of ULF
    if mood == MOOD_DECL:
      lf = [EXISTS, u, restrictor, [[SPEAKER, TELL, HEARER, [THAT, embedded]], CH, u]]
    elif mood == MOOD_QUES:
      lf = [EXISTS, u, restrictor, [[SPEAKER, ASK, HEARER, [ANS_TO, embedded]], CH, u]]
    elif mood == MOOD_IMPER:
      lf = [EXISTS, u, restrictor, [[SPEAKER, INSTRUCT, HEARER, [KA, embedded]], CH, u]]

    return lf


"""Tense node implementation"""

from ulflib.tense_tree.constants import *

class TenseNode:
  """A node in a tense tree.
  
  Attributes
  ----------
  left : TenseNode
    The 'past' branch of this node.
  down : TenseNode
    The 'perfect' branch of this node.
  right : TenseNode
    The 'future' branch of this node.
  embedded : list[TenseNode]
    A list of embedded nodes.
  parent : TenseNode
    The parent of this node.
  embedding : TenseNode
    The node embedding this node.
  episodes : list[str]
    The stack of episode symbols corresponding to this node.
  past_dominated : bool
    Whether this node is part of a past-dominated branch.
  """

  def __init__(self, left=None, down=None, right=None, embedded=[], parent=None, embedding=None, episodes=[], past_dominated=False):
    self.left = left
    self.down = down
    self.right = right
    self.embedded = embedded
    self.parent = parent
    self.embedding = embedding
    self.episodes = episodes
    self.past_dominated = past_dominated

  def move_left(self):
    """Move focus to the 'left' (past tense) child, creating a new node if one doesn't exist."""
    if self.left is None:
      self.left = TenseNode(parent=self, past_dominated=True)
    return self.left
    
  def move_down(self):
    """Move focus to the 'down' (perfect tense) child, creating a new node if one doesn't exist."""
    if self.down is None:
      self.down = TenseNode(parent=self, past_dominated=self.past_dominated)
    return self.down
    
  def move_right(self):
    """Move focus to the 'right' (future tense) child, creating a new node if one doesn't exist."""
    if self.right is None:
      self.right = TenseNode(parent=self, past_dominated=self.past_dominated)
    return self.right

  def move_embedded_cur(self):
    """Move focus into an existing embedded child, creating a new node if one doesn't exist."""
    if not self.embedded:
      self.embedded.append(TenseNode(embedding=self, past_dominated=self.past_dominated))
    return self.embedded[-1]
  
  def move_embedded_new(self):
    """Move focus into a new embedded child; create a new node regardless of whether there is an existing one."""
    self.embedded.append(TenseNode(embedding=self, past_dominated=self.past_dominated))
    return self.embedded[-1]

  def move_parent(self):
    """Move focus into the parent."""
    return self.parent
  
  def move_embedding(self):
    """Move focus into the node embedding this node."""
    return self.embedding

  def add_episode(self, ep):
    """Add an episode to this node."""
    self.episodes.append(ep)
    return self
  
  def get_episodes(self):
    """Get all the episodes of this node."""
    return self.episodes
  
  def last_episode(self):
    """Get the last episode of this node."""
    return self.episodes[-1] if self.episodes else None
  
  def bef(self):
    """This is 'before' if not past-dominated, otherwise 'at-or-before'."""
    return AT_OR_BEFORE if self.past_dominated else BEFORE


import sys
sys.path.append("src/")

import pytest

from ulflib.util import variablep
from ulflib.tense_tree.preprocess import *
from ulflib.tense_tree import TenseTree

def equal_with_variables(ulf1, ulf2):
  var_map = {}
  def rec(x1, x2):
    if atom(x1) and atom(x2):
      if variablep(x1) and variablep(x2):
        if x1 not in var_map:
          var_map[x1] = x2
          return True
        else:
          return var_map[x1] == x2
      else:
        return x1 == x2
    elif atom(x1) or atom(x2):
      return False
    elif len(x1) != len(x2):
      return False
    else:
      return all([rec(y1, y2) for y1, y2 in zip(x1, x2)])
  return rec(ulf1, ulf2)



class TestPreprocessingUninvertAux:
  def test_1(self):
    ulf = [[['pres', 'do.aux-s'], 'john', ['go.v', ['to', ['the.d', 'store.n']]]], '?']
    ulf1 = [['john', [['pres', 'do.aux-s'], ['go.v', ['to', ['the.d', 'store.n']]]]], '?']
    assert equal_with_variables(uninvert_aux(ulf), ulf1)



class TestPreprocessingLiftSentMods:
  def test_1(self):
    ulf = ['past', ['mary', ['leave.v', 'yesterday.adv-e']]]
    ulf1 = ['past', ['yesterday.adv-e', ['mary', 'leave.v']]]
    assert equal_with_variables(lift_sent_mods(ulf), ulf1)


  def test_2(self):
    ulf = ['past', ['mary', ['leave.v', 'yesterday.adv-e', ['adv-e', ['before.p', ['k', 'noon.n']]]]]]
    ulf1 = ['past', ['yesterday.adv-e', [['adv-e', ['before.p', ['k', 'noon.n']]], ['mary', 'leave.v']]]]
    assert equal_with_variables(lift_sent_mods(ulf), ulf1)


  def test_3(self):
    ulf = ['past', ['john', ['go.v', ['to.p', ['the.d', 'bathroom.n']], 'frequently.adv-f', 'yesterday.adv-e']]]
    ulf1 = ['past', ['frequently.adv-f', ['yesterday.adv-e', ['john', ['go.v', ['to.p', ['the.d', 'bathroom.n']]]]]]]
    assert equal_with_variables(lift_sent_mods(ulf), ulf1)


  def test_4(self):
    ulf = ['past', [['the.d', ['red.a', 'block.n']], [['pasv', 'move.v'], ['adv-e', ['most.mod-a', ['adv-e', 'recent.a']]]]]]
    ulf1 = ['past', [['adv-e', ['most.mod-a', ['adv-e', 'recent.a']]], [['the.d', ['red.a', 'block.n']], ['pasv', 'move.v']]]]
    assert equal_with_variables(lift_sent_mods(ulf), ulf1)


  def test_5(self):
    ulf = ['past', ['john', 'just.adv-e', ['move.v', ['the.d', ['green.a', 'block.n']],
                                           ['before.ps', ['past', [['the.d', ['red.a', 'block.n']], [['pasv', 'move.v'],
                                                                      ['adv-e', ['most.mod-a', ['adv-e', 'recent.a']]]]]]]]]]
    ulf1 = ['past', ['just.adv-e', [['before.ps', ['past', [['adv-e', ['most.mod-a', ['adv-e', 'recent.a']]],
                                                            [['the.d', ['red.a', 'block.n']], ['pasv', 'move.v']]]]],
                                    ['john', ['move.v', ['the.d', ['green.a', 'block.n']]]]]]]
    assert equal_with_variables(lift_sent_mods(ulf), ulf1)



class TestPreprocessingExpandTemporalMods:
  def test_ever(self):
    ulf = ['past', ['ever.adv-e', ['john', ['go.v', ['to.p', ['the.d', 'store.n']]]]]]
    ulf1 = ['past', [['adv-e', 'episode.n'], ['john', ['go.v', ['to.p', ['the.d', 'store.n']]]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_indexical(self):
    ulf = ['past', ['yesterday.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', '^yesterday']], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['now.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', '^now']], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['currently.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', '^now']], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_temporal_location(self):
    ulf = ['past', [['adv-e', ['on.p', ['k', ['last.a', 'saturday.n']]]], ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', ['the', '?t', ['?t', ['last.a', 'saturday.n']]]]], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['just.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', ['the', '?t', ['?t', ['previous.a', 'episode.n']]]]], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['initially.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', ['the', '?t', ['?t', ['initial.a', 'episode.n']]]]], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-e', ['before.p', ['this.d', 'wednesday.n']]],
                    [['adv-e', ['after.p', ['the.d', ['last.a', 'weekend.n']]]], ['mary', 'leave.v']]]]
    ulf1 = ['past', [['adv-e', ['before.p', ['this', '?t1', ['?t1', 'wednesday.n']]]],
                     [['adv-e', ['after.p', ['the', '?t2', ['?t2', ['last.a', 'weekend.n']]]]], ['mary', 'leave.v']]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

  
  def test_temporal_range(self):
    ulf = ['past', ['recently.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', ['some', '?t', ['?t', ['recent.a', 'episode.n']]]]], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['before.adv-e', ['mary', 'leave.v']]]
    ulf1 = ['past', [['adv-e', ['during.p', ['some', '?t', ['?t', ['previous.a', 'episode.n']]]]], ['mary', 'leave.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-e', [['mod-a', ['by.p', ['1.d', 'move.n']]], 'ago.a']],
                    [['which.d', ['plur', 'block.n']], ['be.v', ['near.p', ['the.d', ['mercedes', 'block.n']]]]]]]
    ulf1 = ['past', [['adv-e', ['during.p', ['some', '?t', ['?t', [[['mod-a', ['by.p', ['1.d', 'move.n']]], 'ago.a'], 'episode.n']]]]],
                    [['which.d', ['plur', 'block.n']], ['be.v', ['near.p', ['the.d', ['mercedes', 'block.n']]]]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_duration(self):
    ulf = ['past', [['adv-e', ['for.p', ['two.d', ['plur', 'hour.n']]]], ['john', 'sleep.v']]]
    ulf1 = ['past', [['adv-e', ['has-duration.p', ['k', ['2.mod-n', ['plur', 'hour.n']]]]], ['john', 'sleep.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-e', ['in.p', ['two.d', ['plur', 'hour.n']]]], ['john', ['run.v', ['the.d', 'race.n']]]]]
    ulf1 = ['past', [['adv-e', ['in-span-of.p', ['k', ['2.mod-n', ['plur', 'hour.n']]]]], ['john', ['run.v', ['the.d', 'race.n']]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_count(self):
    ulf = ['past', ['twice.adv-f', ['john', ['see.v', ['the.d', 'movie.n']]]]]
    ulf1 = ['past', [['adv-f', ['2.mod-n', ['plur', 'episode.n']]], ['john', ['see.v', ['the.d', 'movie.n']]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-f', ['for.p', ['three.d', ['plur', 'time.n']]]], ['mary', ['visit.v', 'paris']]]]
    ulf1 = ['past', [['adv-f', ['3.mod-n', ['plur', 'time.n']]], ['mary', ['visit.v', 'paris']]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-f', 'twice.a'], ['mary', ['visit.v', 'paris']]]]
    ulf1 = ['past', [['adv-f', ['2.mod-n', ['plur', 'episode.n']]], ['mary', ['visit.v', 'paris']]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_frequency(self):
    ulf = ['past', ['frequently.adv-f', ['john', ['call.v', 'mary']]]]
    ulf1 = ['past', [['adv-f', [['attr', 'frequent.a'], ['plur', 'episode.n']]], ['john', ['call.v', 'mary']]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', ['rarely.adv-f', ['john', ['call.v', 'mary']]]]
    ulf1 = ['past', [['adv-f', [['attr', 'infrequent.a'], ['plur', 'episode.n']]], ['john', ['call.v', 'mary']]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_periodicity(self):
    ulf = ['past', [['adv-f', ['for.p', ['every.d', ['four.a', ['plur', 'hour.n']]]]], ['john', ['take.v', ['k', 'medicine.n']]]]]
    ulf1 = ['past', [['adv-f', [':l', '?s', [['?s', [['attr', 'periodic.a'], ['plur', 'episode.n']]], 'and.cc',
                                             [['period-of.f', '?s'], '=', ['k', ['4.mod-n', ['plur', 'hour.n']]]]]]],
                                             ['john', ['take.v', ['k', 'medicine.n']]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_recurrence(self):
    ulf = ['pres', [['adv-f', ['for.p', ['every.d', 'saturday.n']]], ['mary', 'swim.v']]]
    ulf1 = ['pres', [['adv-f', [':l', '?s', ['all', '?t', ['?t', 'saturday.n'],
                                             ['exists', '?e', [['?e', 'member-of.p', '?s'], 'and.cc',
                                                               ['?e', 'during.p', '?t']]]]]], ['mary', 'swim.v']]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  # def test_temporal_relation(self):
  #   # TODO: unsure if this needs to be preprocessed
  #   ulf = ['past', [['since.ps', ['past', ['john', 'go_off.v', ['to.p', ['k', 'college.n']]]]], ['mary', 'take_care_of.v', ['the.d', 'dog.n']]]]
  #   ulf1 = ['past', [['adv-e', [':l', '?e', ['the-earliest', '?t',
  #                                            ['past', [['adv-e', ['at-time.p', '?t']], ['john', 'go_off.v', ['to.p', ['k', 'college.n']]]]],
  #                                            ['?e', 'after.p', '?t']]]],
  #                       ['mary', 'take_care_of.v', ['the.d', 'dog.n']]]]
  #   assert equal_with_variables(expand_temporal_mods(ulf), ulf1)


  def test_combinations(self):
    ulf = ['past', [['adv-e', ['in.p', ['two.d', ['plur', 'month.n']]]],
                    [['adv-f', ['for.p', ['three.d', ['plur', 'time.n']]]], ['mary', ['visit.v', 'paris']]]]]
    ulf1 = ['past', [['adv-e', ['in-span-of.p', ['k', ['2.mod-n', ['plur', 'month.n']]]]],
                    [['adv-f', ['3.mod-n', ['plur', 'time.n']]], ['mary', ['visit.v', 'paris']]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)

    ulf = ['past', [['adv-e', ['for.p', ['ten.d', ['plur', 'day.n']]]],
                    [['adv-f', ['for.p', ['every.d', ['four.a', ['plur', 'hour.n']]]]], ['john', ['take.v', ['k', 'medicine.n']]]]]]
    ulf1 = ['past', [['adv-e', ['has-duration.p', ['k', ['10.mod-n', ['plur', 'day.n']]]]],
                    [['adv-f', [':l', '?s', [['?s', [['attr', 'periodic.a'], ['plur', 'episode.n']]], 'and.cc',
                                             [['period-of.f', '?s'], '=', ['k', ['4.mod-n', ['plur', 'hour.n']]]]]]],
                        ['john', ['take.v', ['k', 'medicine.n']]]]]]
    assert equal_with_variables(expand_temporal_mods(ulf), ulf1)



tree = TenseTree(disable_speech_acts=True)
# ulf = ['mary', [['past', 'leave.v'], ['adv-e', ['before.p', ['this.d', 'wednesday.n']]]]]
ulf = ['past', ['frequently.adv-f', ['john', ['call.v', 'mary']]]]
print(tree.deindex(ulf))
# print(preprocess(ulf))
# print(scoping.scope(ulf))




"""Functions for scoping a ULF formula."""

from transduction import tt
from memoization import cached

from ulflib.util import atom, variablep, listp, append, occurs_in, occurs_properly_in, subst, rec_remove, flatten_singletons, new_var
from ulflib.ulflib import (det_term_p, coord_p, coord_sent_p, tensed_sent_p, nonsubsective_premodified_verb_p, sent_p, pred_p, verbal_sent_p,
                           neg_non_verbal_sent_p, neg_sent_p, non_neg_modified_sent_p, tense_or_aspect_marker_p, tensed_verb_p,
                           tensed_sent_reifier_p, lex_ps_p,
                           premodified_verbal_sent_p, sent_or_tensed_sent_p, lambda_p, lex_function_p, find_vp_head, split_by_suffix)

# ``````````````````````````````````````
# TT Scoped LF Patterns
# ``````````````````````````````````````



QUANTIFIER = ['all', 'some', 'a', 'a{n}', 'most', 'few', 'many', 'the', 'this', 'these', 'those']
DEFAULT_TYPES = ['tense', 'quan', 'coord']


@cached
def scoped_tensed_sent_p(x):
  return listp(x) and len(x) == 2 and tense_or_aspect_marker_p(x[0]) and sent_p(x[1])


@cached
def scoped_quant_sent_p(x):
  return listp(x) and ((len(x) == 3 and x[0] in QUANTIFIER and (sent_p(x[2]) or tensed_sent_p(x[2]))) or
                       (len(x) == 4 and x[0] in QUANTIFIER and (sent_p(x[2]) or tensed_sent_p(x[2])) and
                        (sent_p(x[3]) or tensed_sent_p(x[3]))))


SCOPED_PREDS = [
  scoped_tensed_sent_p,
  scoped_quant_sent_p,
]

for pred in SCOPED_PREDS:
  tt.register_pred(pred, include_neg=True)



# ``````````````````````````````````````
# Helper Functions
# ``````````````````````````````````````



def occurs_free_in(atm, expr, bound_vars=[]):
  """Check if `atm` occurs in `expr`, without being bound therein by lambda or a scoped quantifier.
  
  The optional `bound_vars` are gathered as we work our way downward to the parts of `expr` recursively.
  """
  if atm in bound_vars:
    return False
  elif atom(expr):
    return True if atm == expr else False
  else:
    bound_vars1 = bound_vars
    if lambda_p(expr) or (expr[0] in QUANTIFIER and atom(expr[1])):
      bound_vars1.append(expr[1])
    return any([occurs_free_in(atm, xpr, bound_vars1) for xpr in expr])



def subst_for_free(val, x, expr):
  """Substitute `val` for all free occurrences of atom `x` in `expr`."""
  if expr == x:
    return val
  elif atom(expr):
    return expr
  elif not occurs_in(x, expr):
    return expr
  # is x bound at the top of expr by a quantifier or :l?
  elif len(expr) >= 3 and variablep(expr[1]) and not lex_function_p(expr[0]):
    # if the bound variable is x, return expr unchanged
    if expr[1] == x:
      return expr
    # if occurs-check is positive, first rename the variable and then substitute recursively
    elif occurs_in(expr[1], val):
      new = new_var([val, expr[1:]])
      renamed_expr = expr[0] + subst_for_free(new, expr[1], expr[1:])
      return [subst_for_free(val, x, y) for y in renamed_expr]
    # the bound variable isn't x, and doesn't occur in val
    else:
      return [subst_for_free(val, x, y) for y in expr]
  # x occurs within expr, but not as a top-level bound variable
  else:
    return [subst_for_free(val, x, y) for y in expr]
  


def to_quantifier(det):
  """Convert a determiner to a particular quantifier (if possible)."""
  if not atom(det):
    return det
  if det in ['a.d', 'an.d']:
    return 'a{n}'
  word, _ = split_by_suffix(det)
  if word and word in QUANTIFIER:
    return word
  else:
    return det

 

# ``````````````````````````````````````
# Scoping Algorithm
# ``````````````````````````````````````



def scope(ulf, types=DEFAULT_TYPES):
  """The top-level function for scoping a ULF expression.
  
  Parameters
  ----------
  ulf : s-expr
    An "S-expression" (possibly nested lists of strings) representing an
    unscoped logical form (ULF) formula.
  types : list[str], optional
    The types of categories to scope; currently supported are
    ``tense``, ``quan``, and ``coord``.
  
  Returns
  -------
  s-expr
    The scoped logical form.
  """
  candidates = scoping_candidates(ulf)

  if not candidates:
    return [scope_expr(part, types=types) for part in ulf]
  
  expr = wide_scope_winner(candidates)

  if coord_p(expr) and 'coord' in types:
    return scope_coord(expr, ulf, types=types)
  elif det_term_p(expr) and 'quan' in types:
    return scope_quan(expr, ulf, types=types)
  # TODO: should this match just (past pred.v), or the full tensed sentence?
  elif tensed_sent_p(expr) and 'tense' in types:
    return scope_tense(expr, ulf, types=types)
  elif 'coord' in types:
    return scope_coord(expr, ulf, types=types)
  else:
    return [scope_expr(part, types=types) for part in ulf]
  

def scope_expr(ulf, types=DEFAULT_TYPES):
  """Scope an arbitrary ULF expression.
  
  The input `ulf` could be any expression, but we only scope sentences and monadic preds
  (though wff/pred *parts* of other expressions are also scoped by recursion).
  """
  if atom(ulf):
    return ulf
  elif sent_or_tensed_sent_p(ulf):
    return scope(ulf, types=types)
  elif pred_p(ulf):
    return scope_pred(ulf, types=types)
  else:
    return [scope_expr(part, types=types) for part in ulf]
  

def scoping_candidates(ulf):
  """Find accessible unscoped elements, if any, in the given expr.

  Return the expr itself if it is unscoped, followed by the scoping candidates from
  all parts that are *not* scope islands as determined by the island constraints listed in README.md.
  """
  candidates = []

  if atom(ulf):
    return []
  
  # Per island constraints 2 and 4: skip sentence modifiers other than 'not' or already scoped expressions
  if non_neg_modified_sent_p(ulf) or scoped_quant_sent_p(ulf) or scoped_tensed_sent_p(ulf):
    return []
  
  # if ulf consists of 'not' applied to a NONverbal wff, return empty
  # (normally only tense can escape from a negation)
  if neg_non_verbal_sent_p(ulf):
    return []
  
  # if ulf consists of 'not' applied to a wff that has an external unscoped tense operator
  # applied to it, extract the tense expression and return it in a singleton list
  if neg_sent_p(ulf):
    tensed_expr = find_unscoped_tense_expr(ulf[1])
    if tensed_expr:
      return [tensed_expr]
    # failure does not mean there's no unscoped tense; it may be part of the predicate of the wff
  
  # Per island constraint 3: if ulf consists of a nonsubsective predicate modifier applied to a predicate,
  # extract the tense expression of the predicate (if any) and return it as a singleton list
  if nonsubsective_premodified_verb_p(ulf):
    tensed_expr = find_unscoped_tense_expr(ulf[1])
    if tensed_expr:
       return [tensed_expr]
    else:
       return []
    
  # Per island constraint 5: if ulf is an unscoped coordination of wffs, extract only the ulf itself
  if coord_sent_p(ulf):
    return [ulf]
  
  # if ulf is an unscoped quantified term, an unscoped tense-modified expression,
  # or an unscoped coordinated expression, then initiate the candidates with [ulf]
  # TODO: should this match just (past pred.v), or the full tensed sentence?
  if det_term_p(ulf) or tensed_sent_p(ulf) or coord_p(ulf):
    candidates = [ulf]

  # recursively add further candidates
  # Per island constraint 6: if a sub-expression is a verbal wff, we do not process it
  candidates = candidates + append([scoping_candidates(expr) if not (atom(expr) or verbal_sent_p(expr)) else [] for expr in ulf])

  # Per island constraint 1: drop candidates that contain a free variable bound in ulf outside of the candidate
  candidates = [c for c in candidates if not binding_constrained_candidate(c, ulf)]

  return candidates


def binding_constrained_candidate(xp, expr):
  """Determine whether `xp` contains a free variable bound in `expr` outside of itself.
  
  If there is a variable-binding operator at the top level, and the variable it binds occurs
  free in `xp`, return True; otherwise check for each top-level subexpression of `expr` whether
  `xp` occurs in it and is binding-constrained relative to it, returning True if this occurs for
  any of those subexpressions, and False otherwise.
  """
  if atom(xp) or atom(expr):
    return False
  if xp == expr:
    return False
  if not occurs_properly_in(xp, expr):
    return False
  
  # top-level lambda, or top-level scoped quantifier
  if ((lambda_p(expr) and occurs_free_in(expr[1], xp)) or
      (expr[0] in QUANTIFIER and atom(expr[1]) and occurs_free_in(expr[1], xp))):
    return True
  
  # embedded binding operator, outside xp, binding something in xp
  return any([binding_constrained_candidate(xp, xpr) for xpr in expr])
    

def find_unscoped_tense_expr(verbal_pred):
  """Find a top-level or superficially embedded unscoped-tense expression, given a verbal predication.
  
  Usually this expression is just a tensed atomic verb, but some might have non-atomic logical translations.

  This program is necessary because tense tends to scope outside modified verbal predicates *even when the modifier
  is non-subsective* (whereas quantifiers and binary connectives tend to stay inside). For example, in "He almost died",
  it's not that an event of almost-dying-in-the-past happened, but rather, that in the past an event of almost-dying happened.
  In contrast, "He almost solved every problem" is likely to be understood as referring to a near-completion of the entire set
  of problems, rather than a near-completion of each problem individually.
  """
  if atom(verbal_pred):
    return None
  # we do not extract from coordinated expressions
  if coord_p(verbal_pred):
    return None
  # is verbal_pred itself tense-modified?
  if tensed_verb_p(verbal_pred):
    return verbal_pred
  # is verbal_pred a verbal pred applied to at least one argument?
  if verbal_sent_p(verbal_pred):
    vp = verbal_pred[1]
    head = find_vp_head(vp)
    return head if head else find_unscoped_tense_expr(verbal_pred[1])
  # is verbal_pred a modified verbal pred?
  if premodified_verbal_sent_p(verbal_pred):
    return find_unscoped_tense_expr(verbal_pred[1])
  # lambda abstract forming a verbal pred?
  if lambda_p(verbal_pred):
    lexpr = verbal_pred[2]
    # is this an abstraction from another verbal pred or an uncoordinated wff?
    if tensed_verb_p(lexpr) or (verbal_sent_p(lexpr) and not coord_p(lexpr)):
      return find_unscoped_tense_expr(lexpr)
  return None
      

def wide_scope_winner(unscoped_elements):
  """Pick the widest-scoping unscoped element in `ulf`.
  
  The given unscoped-elements are in left-to-right order, and
  (justified by the constituent order in unscoped LF) we pick 
  the leftmost as the default, except for giving preference to tense.

  Notes
  -----
  TODO: This function may need to be tweaked, as it's currently taken from the
  original elf-from-sentences code and the ULF tense syntax is slightly different.
  """
  if not unscoped_elements:
    return []
  tense_expr = [x for x in unscoped_elements if tense_or_aspect_marker_p(x)]
  if tense_expr:
    return tense_expr[0]
  else:
    return unscoped_elements[0]
  

def extract_tense(ulf):
  """Extract the tense marker from a ulf (blocking on sentence-embedding operators)."""
  if tense_or_aspect_marker_p(ulf):
    return ulf
  elif atom(ulf):
    return None
  elif tensed_sent_reifier_p(ulf[0]) or lex_ps_p(ulf[0]):
    return None
  else:
    part1 = extract_tense(ulf[0])
    if part1:
      return part1
    else:
      return extract_tense(ulf[1:])
    

def remove_tense(tense, ulf):
  """Recursively remove a given tense marker from a ulf (blocking on sentence-embedding operators)."""
  new_lst = []
  for e in ulf:
    if e == tense:
      continue
    elif type(e) == list and not (tensed_sent_reifier_p(e[0]) or lex_ps_p(e[0])):
      new_lst.append(remove_tense(tense, e))
    else:
      new_lst.append(e)

  return new_lst
  

def scope_pred(pred, types=DEFAULT_TYPES):
  """Used when no unscoped elements of `pred` can be scoped as part of a larger ULF.
  
  NOTES
  -----
  TODO: we could first check if pred contains any unscoped elements, to avoid reassembling
  pred in that case, but this is not so simple to check, so is not included yet.
  """
  if atom(pred):
    return pred
  
  # Modified pred?
  if len(pred) == 2 and nonsubsective_premodified_verb_p(pred[0]) and pred_p(pred[1]):
    return [scope_expr(pred[0], types=types), scope_pred(pred[1], types=types)]

  # Lambda predicate?
  if lambda_p(pred):
    return [pred[0], pred[1], scope(pred[2], types=types)]

  # Nonmonadic pred applied to term(s)?
  # TODO: the following was copied from the original scoping method, but it appears to create
  # an infinite recursion. I'm not entirely sure what the intended purpose of the lambda is.
  # if len(pred) >= 2 and pred_p(pred):
  #   var = new_var(pred)
  #   return [':l', var, scope([var, pred], types=types)]
  # TODO: the following needs to be carefully validated
  if listp(pred) and len(pred) >= 2 and pred_p(pred):
    inner = scope(pred[1:], types=types)
    if scoped_quant_sent_p(inner):
      return [pred[0], inner]
    return [pred[0]] + inner

  return [scope_expr(p, types=types) for p in pred]


def scope_quan(expr, ulf, types=DEFAULT_TYPES):
  """Scope a ULF of form ``[<quan>, <pred>]``."""
  quan = to_quantifier(expr[0])
  x = new_var(ulf)
  pred = expr[1]
  # Prepare restrictor using new variable as subject
  if atom(pred):
    restr = [x, pred]
  elif lambda_p(pred):
    restr = subst_for_free(x, pred[1], pred[2])
  elif pred_p(pred):
    restr = [x] + pred
  else:
    restr = [x, pred]
  # Is restr actually the body of (unrestricted) ulf?
  if expr == ulf:
    return [quan, x, scope(restr, types=types)]
  # ulf is restricted
  else:
    return [quan, x, scope(restr, types=types), scope(subst(x, expr, ulf), types=types)]


def scope_tense(expr, ulf, types=DEFAULT_TYPES):
  """Scope a ULF of form ``[<tense>, <expr>]``."""
  tense = extract_tense(expr)
  if not tense: # unexpected
    return ulf
  sent = flatten_singletons(remove_tense(tense, expr))
  if expr == ulf:
    return scope([tense, sent], types=types)
  else:
    return [tense, scope(subst(sent, expr, ulf), types=types)]


def scope_coord(expr, ulf, types=DEFAULT_TYPES):
  """Scope a ULF of form ``[<expr1>, <coord>, <expr2>, ...]``."""
  # Top-level sentence coordination
  if expr == ulf:
    return [scope(expr[0], types=types), expr[1]] + [scope(xp, types=types) for xp in expr[2:]]
  else:
    return [scope(subst(expr[0], expr, ulf), types=types), expr[1]] + [scope(subst(xp, expr, ulf), types=types) for xp in expr[2:]]


"""Word lists and word->predicate mappings used in adverbial expansion."""

DAYS = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday']
TIMES_OF_DAY = ['dawn', 'sunrise', 'morning', 'noon', 'afternoon', 'evening', 'sundown', 'sunset', 'dusk', 'night', 'midnight']
INDEXICAL_TIMES = ['yesterday', 'tomorrow', 'now', 'current'] + TIMES_OF_DAY + DAYS
ABSOLUTE_TIMES = ['second', 'minute', 'hour', 'day', 'month', 'year', 'decade', 'century']
NOUNS = DAYS + TIMES_OF_DAY + ABSOLUTE_TIMES + ['weekend', 'weekday']

NUMBERS = {'one':1, 'two':2, 'three':3, 'four':4, 'five':5, 'six':6, 'seven':7, 'eight':8, 'nine':9, 'ten':10,
           'eleven':11, 'twelve':12, 'thirteen':13, 'fourteen':14, 'fifteen':15, 'sixteen':16, 'seventeen':17,
           'eighteen':18, 'nineteen':19, 'twenty':20, 'thirty':30, 'forty':40, 'fifty':50, 'sixty':60, 'seventy':70,
           'eighty':80, 'ninety':90, 'one_hundred':100, 'hundred':100, 'thousand':1000,
           'once':1, 'twice':2, 'thrice':3,
           'one-time':1, 'two-time':2, 'three-time':3, 'four-time':4, 'five-time':5}

LOCATION_PRED_MAP = {
  'during': 'during.p', 'at': 'during.p', 'in': 'during.p', 'within': 'during.p', 'on': 'during.p',
  'before': 'before.p', 'prior_to': 'before.p', 'preceding': 'before.p', 'until': 'before.p',
  'after': 'after.p', 'following': 'after.p', 'since': 'after.p', 'from': 'after.p'
}

LOCATION_ADV_MAP = {
  'originally': 'initial.a', 'initially': 'initial.a',
  'last': 'last.a', 'finally': 'last.a',
  'just': 'previous.a',
  'first': 'first.a', 'second': 'second.a', 'third': 'third.a', 'fourth': 'fourth.a', 'fifth': 'fifth.a',
  'sixth': 'sixth.a', 'seventh': 'seventh.a', 'eight': 'eight.a', 'nineth': 'nineth.a', 'tenth': 'tenth.a'
}

LOCATION_ADJ_MAP = {
  'original': 'initial.a', 'initial': 'initial.a',
  'last': 'last.a', 'final': 'last.a',
  'just': 'previous.a',
  'first': 'first.a', 'second': 'second.a', 'third': 'third.a', 'fourth': 'fourth.a', 'fifth': 'fifth.a',
  'sixth': 'sixth.a', 'seventh': 'seventh.a', 'eight': 'eight.a', 'nineth': 'nineth.a', 'tenth': 'tenth.a'
}

RANGE_ADV_MAP = {
  'recently': 'recent.a',
  'previously': 'previous.a', 'before': 'previous.a',
  'next': 'next.a', 'later': 'next.a',
}

RANGE_ADJ_MAP = {
  'recent': 'recent.a',
  'previous': 'previous.a', 'before': 'previous.a', 'preceding': 'previous.a',
  'next': 'next.a', 'later': 'next.a', 'following': 'next.a', 'future': 'next.a',
}

DURATION_PRED_MAP = {
  'for': 'has-duration.p', 'during': 'has-duration.p',
  'in': 'in-span-of.p'
}

FREQUENCY_ADV_MAP = {
  'frequently': 'frequent.a', 'commonly': 'frequent.a', 'regularly': 'frequent.a', 'often': 'frequent.a', 'routinely': 'frequent.a',
  'infrequently': 'infrequent.a', 'rarely': 'infrequent.a', 'uncommonly': 'infrequent.a', 'irregularly.a': 'infrequent.a', 'seldom': 'infrequent.a',
  'occasionally': 'infrequent.a', 'sometimes': 'infrequent.a'
}

FREQUENCY_ADJ_MAP = {
  'frequent': 'frequent.a', 'common': 'frequent.a', 'regular': 'frequent.a', 'often': 'frequent.a', 'routine': 'frequent.a',
  'infrequent': 'infrequent.a', 'rare': 'infrequent.a', 'uncommon': 'infrequent.a', 'irregular.a': 'infrequent.a', 'seldom': 'infrequent.a',
  'occasional': 'infrequent.a', 'sometimes': 'infrequent.a'
}

FREQUENCY_DET_MAP = {
  'many': 'frequent.a', 'most': 'frequent.a',
  'some': 'infrequent.a', 'few': 'infrequent.a'
}

RECURRENCE_ADV_MAP = {
  'always': 'episode.n', 'constantly': 'episode.n'
}

RECURRENCE_ADJ_MAP = {
  'always': 'episode.n', 'constant': 'episode.n'
}

ADV_F_P = ['for', 'during', 'on', 'at']


QUAN_MAP = {
  'k': 'the',
  'the.d': 'the',
  'this.d': 'this'
}


NOW = 'Now'
EP_VAR = '?E'
UTT_VAR = '?U'

MOOD_DECL = 'decl'
MOOD_QUES = 'ques'
MOOD_IMPER = 'imper'

SPEAKER = '^Speaker'
HEARER = '^Hearer'

EXISTS = 'exists'
AND = 'and.cc'
CH = '**'
MULT = 'mult'

SAME_TIME = 'same-time.p'
BEFORE = 'before.p'
AFTER = 'after.p'
AT_OR_BEFORE = 'at-or-before.p'
JUST_AFTER = 'just-after.p'
AT_ABOUT = 'at-about.p'
IMPINGES_ON = 'impinges-on.p'
ORIENTS = 'orients.p'

TELL = 'tell.v'
ASK = 'ask.v'
INSTRUCT = 'instruct.v'

THAT = 'that'
ANS_TO = 'ans-to'
KA = 'ka'

PAST = 'past'
PRES = 'pres'
FUTR = 'futr'
PERF = 'perf'


"""Utilities for handling S-expression lists."""

from ulflib.util import atom, symbolp, escaped_symbol_p, isquote, replaceall

def balanced_substr(s):
	"""Find a substring with a balanced number of parentheses."""
	count = 1
	for i in range(1, len(s)):
		c = s[i]

		if c == '(':
			count += 1
		if c == ')':
			count -= 1
		if count == 0:
			return s[:i+1]

	return None


def clean_s_expr(s_expr):
	"""Clean a string representation of an S-expression by removing newlines and standardizing whitespace."""
	s_expr = s_expr.replace('\n', '')
	s_expr = s_expr.replace('\t', '')
	s_expr = replaceall(s_expr,
						[(r"\| ([a-zA-Z0-9-_.']+)\|", r'|\1|', True),
						 (r"\|([a-zA-Z0-9-_.']+) \|", r'|\1|', True),
						 # For handling proper names with multiple words (up to 4)
						 (r"\|([a-zA-Z0-9-_.']+)[ ]?([a-zA-Z0-9-_.']+)?[ ]?([a-zA-Z0-9-_.']+)?[ ]?([a-zA-Z0-9-_.']+)?\|",
							r'|\1_\2_\3_\4|', True)])
	while '__' in s_expr:
		s_expr = s_expr.replace('__', '_')
	while '  ' in s_expr:
		s_expr = s_expr.replace('  ', ' ')
	return s_expr


def standardize_symbols(s_expr):
	"""Standardize the symbols within an S-expression by mapping to lowercase, unless enclosed in escape symbols."""
	def standardize_rec(e):
		if symbolp(e) and not escaped_symbol_p(e) and not isquote(e):
			return e.lower()
		elif symbolp(e) and escaped_symbol_p(e):
			parts = e.split('|')
			before = parts[0].lower()
			escaped = parts[1]
			after = parts[2].lower()
			if '.' in escaped:
				word, suffix = escaped.split('.')
				if suffix and suffix[0] != '_':
					escaped = word
					suffix = suffix.strip('_')
					after = '.'+suffix.lower()+after
			escaped = escaped.strip('_').replace('_', ' ')
			return before+escaped+after
		elif symbolp(e) and isquote(e):
			return e
		else:
			return [standardize_rec(x) for x in e]
	return standardize_rec(s_expr)


def convert_quotes(s_expr):
	"""Convert any quoted word lists (i.e., single ' symbol followed by a list of symbols) to a single quoted string."""
	def convert_quotes_rec(e):
		if atom(e):
			return e
		elif len(e) == 1:
			return [convert_quotes_rec(e[0])]
		else:
			e1 = []
			for x1, x2 in zip([None]+e[:-1], e):
				if x1 == "'" and x2 and isinstance(x2, list) and all([isinstance(x, str) for x in x2]):
					e1.append('"'+' '.join(x2)+'"')
				elif x2 != "'":
					e1.append(x2)
			return [convert_quotes_rec(x) for x in e1]
	return convert_quotes_rec(s_expr)


def compress_quotes(s_expr):
	"""Compress quoted expressions that were split between multiple words during parsing."""
	def compress_quotes_rec(e):
		if atom(e):
			return e
		elif len(e) == 1:
			return [compress_quotes_rec(e[0])]
		else:
			e1 = []
			acc = []
			i = 0
			while i < len(e):
				if isinstance(e[i], str) and e[i][0] == '"':
					j = i
					while j < len(e):
						acc.append(e[j])
						if not isinstance(e[j], str):
							e1 += acc
							acc = []
							j += 1
							break
						elif e[j][-1] == '"':
							e1.append(' '.join(acc))
							acc = []
							j += 1
							break
						else:
							j += 1
					i = j
				else:
					e1.append(e[i])
					i += 1
			return [compress_quotes_rec(x) for x in e1]
	return compress_quotes_rec(s_expr)


def parse_s_expr(s_expr):
	"""Parse a string containing an S-expression (in LISP form) into a structured list.
	
	Parameters
	----------
	s_expr : str
		An S-expression in LISP form, e.g., ``(a (b c (d e)) '(f g h))``.
	
	Returns
	-------
	s-expr
		A structured S-expression, i.e., a recursively nested list structure with string "symbols" as atoms.
		e.g., ``['a', ['b', 'c', ['d', 'e']], "f g h"]``
	"""
	def parse_s_expr_rec(s_expr):
		s_expr = clean_s_expr(s_expr)

		if len(s_expr) == 0:
			return None

		if s_expr[0] != '(' or s_expr[-1] != ')':
			return s_expr

		items = []

		item_buf = []
		i = 1
		while i < len(s_expr):
			c = s_expr[i]

			if c == ' ':
				if len(item_buf) > 0:
					items.append(''.join(item_buf))
					item_buf = []
				i += 1
			elif c != '(':
				if c != ')':
					item_buf.append(c)
				i += 1
			else:
				if len(item_buf) > 0:
					items.append(''.join(item_buf))
					item_buf = []
				inner = balanced_substr(s_expr[i:])
				items.append(parse_s_expr_rec(inner))
				i += len(inner)

		if len(item_buf) > 0:
			items.append(''.join(item_buf))
			item_buf = []

		return items

	return convert_quotes(standardize_symbols(compress_quotes(parse_s_expr_rec(s_expr))))